# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import sparse
from chainladder import ARRAY_BACKEND
from sparse import COO as sp
from sparse import elemwise
import pandas as pd
import copy

sp.isnan = np.isnan
sp.newaxis = np.newaxis
sp.nan = np.array([1.0, np.nan])[-1]
sp.testing = np.testing
sp.nansum = sparse.nansum
sp.nanmin = sparse.nanmin
sp.nanmax = sparse.nanmax
sp.concatenate = sparse.concatenate
sp.diagonal = sparse.diagonal
sp.zeros = sparse.zeros
sp.testing.assert_array_equal = np.testing.assert_equal
sp.sqrt = np.sqrt
sp.log = np.log
sp.exp = np.exp
sp.abs = np.abs


def nan_to_num(a):
    if type(a) in [int, float, np.int64, np.float64]:
        return np.nan_to_num(a)
    if hasattr(a, "fill_value"):
        a = a.copy()
        a.data[np.isnan(a.data)] = 0.0
    return sp(coords=a.coords, data=a.data, fill_value=0.0, shape=a.shape)


def ones(*args, **kwargs):
    return sp(np.ones(*args, **kwargs), fill_value=sp.nan)


def nanquantile(a, q, axis=None, keepdims=None, *args, **kwargs):
    new_a = np.nanquantile(
        a.todense(), q=q, axis=axis, keepdims=keepdims, *args, **kwargs
    )
    return sp(np.nan_to_num(new_a), fill_value=np.nan)


def nanmedian(a, axis=None, keepdims=None, *args, **kwargs):
    new_a = np.nanmedian(a.todense(), axis=axis, keepdims=keepdims, *args, **kwargs)
    return sp(np.nan_to_num(new_a), fill_value=np.nan)


def nanmean(a, axis=None, keepdims=None, *args, **kwargs):
    n = sp.nansum(a, axis=axis, keepdims=keepdims)
    d = sp.nansum(sp.nan_to_num(a) != 0, axis=axis, keepdims=keepdims).astype(n.dtype)
    n = sp(data=n.data, coords=n.coords, fill_value=np.nan, shape=n.shape)
    d = sp(data=d.data, coords=d.coords, fill_value=np.nan, shape=d.shape)
    out = n / d
    return sp(data=out.data, coords=out.coords, fill_value=0, shape=out.shape)


def array(a, *args, **kwargs):
    if kwargs.get("fill_value", None) is not None:
        fill_value = kwargs.pop("fill_value")
    else:
        fill_value = sp.nan
    if type(a) == sp:
        return sp(a, *args, **kwargs, fill_value=fill_value)
    else:
        return sp(np.array(a, *args, **kwargs), fill_value=fill_value)


def arange(*args, **kwargs):
    return sparse.COO.from_numpy(np.arange(*args, **kwargs))


def where(*args, **kwargs):
    return elemwise(np.where, *args, **kwargs)


def swapaxes(a, axis1, axis2):
    ax = np.arange(a.ndim)
    axis1 = ax[axis1]
    axis2 = ax[axis2]
    fv = a.fill_value
    l = []
    for item in range(a.ndim):
        if item == axis1:
            l.append(axis2)
        elif item == axis2:
            l.append(axis1)
        else:
            l.append(item)
    coords = a.coords[l, :]
    return sp(
        coords,
        a.data,
        shape=tuple([a.shape[item] for item in l]),
        prune=True,
        fill_value=fv,
    )


def repeat(a, repeats, axis):
    """Repeat elements of an array"""
    ax = np.arange(a.ndim)
    axis = ax[axis]
    r = []
    for item in range(repeats):
        coords = a.coords.copy()
        coords[axis] = coords[axis] + item
        r.append(coords)
    v = np.tile(a.data, repeats)
    a.coords = np.concatenate(r, axis=1)
    a.data = v
    a.shape = tuple(
        [item if num != axis else item * repeats for num, item in enumerate(a.shape)]
    )
    return a


def allclose(a, b, *args, **kwargs):
    return np.allclose(
        np.nan_to_num(a.todense()), np.nan_to_num(b.todense()), *args, **kwargs
    )


def unique(a, *args, **kwargs):
    fv = a.fill_value
    a = np.unique(a.todense(), *args, **kwargs)
    return sp(a, fill_value=fv)


def around(a, *args, **kwargs):
    fv = a.fill_value
    a = np.around(a.todense(), *args, **kwargs)
    return sp(a, fill_value=fv)


def apply_along_axis(func1d, axis, arr, *args, **kwargs):
    fv = arr.fill_value
    arr = np.apply_along_axis(func1d, axis, arr.todense(), *args, **kwargs)
    return sp(arr, fill_value=fv)


def floor(x, *args, **kwargs):
    x.data = np.floor(x.data)
    return x


def minimum(x1, x2):
    if type(x2) in [int, float]:
        out = x1.copy()
        out.data = np.minimum(x1.data, x2)
        return out
    elif np.all(x1.coords == x2.coords):
        out = x1.copy()
        out.data = np.minimum(x1.data, x2.data)
        return out
    if x1.shape != x2.shape:
        raise ValueError("Shapes are not equal")
    return (x1 < x2) * x1 + (x1 >= x2) * x2


def maximum(x1, x2):
    if np.all(x1.coords == x2.coords):
        out = x1.copy()
        out.data = np.maximum(x1.data, x2.data)
        return out
    if x1.shape != x2.shape:
        raise ValueError("Shapes are not equal")
    return (x1 < x2) * x2 + (x1 >= x2) * x1


sp.minimum = minimum
sp.maximum = maximum
sp.unique = unique
sp.around = around
sp.apply_along_axis = apply_along_axis
sp.floor = floor
sp.repeat = repeat
sp.where = where
sp.arange = arange
sp.array = array
sp.nan_to_num = nan_to_num
sp.ones = ones
sp.nanquantile = nanquantile
sp.nanmedian = nanmedian
sp.nanmean = nanmean
sp.swapaxes = swapaxes
sp.allclose = allclose



import chainladder as cl


def test_struhuss():
    X = cl.load_sample("cc_sample")["loss"]
    X = cl.TailConstant(tail=1 / 0.85).fit_transform(cl.Development().fit_transform(X))
    sample_weight = cl.load_sample("cc_sample")["exposure"].latest_diagonal
    ibnr = int(
        cl.CapeCod(trend=0.07, decay=0.75)
        .fit(X, sample_weight=sample_weight)
        .ibnr_.sum()
    )
    assert ibnr == 17052



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
from chainladder.utils import WeightedRegression
from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.development import DevelopmentBase, Development
from chainladder.core.io import EstimatorIO
from chainladder.core.common import Common


class TailBase(BaseEstimator, TransformerMixin, EstimatorIO, Common):
    """ Base class for all tail methods.  Tail objects are equivalent
        to development objects with an additional set of tail statistics"""

    def fit(self, X, y=None, sample_weight=None):
        if X.array_backend == "sparse":
            obj = X.set_backend("numpy")
            xp = np
        else:
            xp = X.get_array_module()
            obj = X.copy()
        if "ldf_" not in obj:
            obj = Development().fit_transform(obj)
        self._ave_period = {"Y": (1, 12), "Q": (4, 3), "M": (12, 1)}[
            obj.development_grain
        ]
        t_ddims = [
            (item + 1) * self._ave_period[1] + obj.ldf_.ddims[-1]
            for item in range(self._ave_period[0])
        ]
        ddims = np.concatenate((obj.ldf_.ddims, t_ddims, np.array([9999])), 0,)
        self.ldf_ = obj.ldf_.copy()
        tail = xp.ones(self.ldf_.shape)[..., -1:]
        tail = xp.repeat(tail, self._ave_period[0] + 1, -1)
        self.ldf_.values = xp.concatenate((self.ldf_.values, tail), -1)
        self.ldf_.ddims = ddims
        if hasattr(obj, "sigma_"):
            zeros = tail[..., 0:obj.ldf_.shape[2], -1:] * 0
            self.sigma_ = getattr(obj, "sigma_").copy()
            self.sigma_.values = xp.concatenate((self.sigma_.values, zeros), -1)
            self.std_err_ = getattr(obj, "std_err_").copy()
            self.std_err_.values = xp.concatenate((self.std_err_.values, zeros), -1)
            self.sigma_.ddims = self.std_err_.ddims = self.ldf_.ddims
            self.sigma_._set_slicers()
            self.std_err_._set_slicers()
        if hasattr(obj, "average_"):
            self.average_ = obj.average_
        else:
            self.average_ = None
        self.ldf_._set_slicers()
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        triangles = ["ldf_", "std_err_", "sigma_"]
        for item in triangles + ["tail_", "_ave_period", "average_"]:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new

    def _get_tail_prediction(self, tail_ldf):
        xp = self.ldf_.get_array_module()
        accum_point = self.ldf_.shape[-1] - 1
        ave = 1 + tail_ldf[..., :accum_point]
        all = xp.prod(1 + tail_ldf[..., accum_point:], -1)[..., None]
        tail = xp.concatenate((ave, all), -1)
        return tail

    def _get_initial_ldf(self, xp, tail):
        """ Quadratic series expansion solution to return seed LDF for tail"""
        arr = self.decay ** xp.arange(1000)
        a = xp.sum(arr ** 2)
        b = xp.sum(arr)
        c = -xp.log(tail)
        return (-b + xp.sqrt(b ** 2 - 4 * a * c)) / (2 * a)

    def _apply_decay(self, X, tail, attach_idx=None):
        """ Created Tail vector with decay over time. """
        xp = self.ldf_.get_array_module()
        if attach_idx:
            decay_range = self.ldf_.shape[-1] - attach_idx
        else:
            decay_range = self.ldf_.shape[-1] - X.shape[-1] + 1
        if xp.max(xp.array(tail)) == 1.0:
            ldfs = 1 + 0 * (self.decay ** xp.arange(1000))
        else:
            ldfs = 1 + self._get_initial_ldf(xp, tail) * (self.decay ** xp.arange(1000))
        ldfs = ldfs[..., :decay_range]
        tail = tail / xp.prod(ldfs[..., :-1], axis=-1, keepdims=True)
        ldfs = xp.concatenate((ldfs[..., :-1], tail), axis=-1)
        self.ldf_.values = xp.concatenate(
            (
                self.ldf_.values[..., :-decay_range],
                (xp.nan_to_num(self.ldf_.values[..., -decay_range:]) * 0 + 1) * ldfs,
            ),
            axis=-1,
        )
        return self

    def _get_tail_stats(self, X):
        """ Method to approximate the tail sigma using
        log-linear extrapolation applied to tail average period
        """
        from chainladder.utils.utility_functions import num_to_nan
        if not hasattr(X, 'sigma_'):
            self.sigma_ = None
            self.std_err_ = None
        else:
            time_pd = self._get_tail_weighted_time_period(X)
            xp = X.sigma_.get_array_module()
            reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(X.sigma_.values), None)
            sigma_ = xp.exp(time_pd * reg.slope_ + reg.intercept_)
            y = X.std_err_.values
            y = num_to_nan(y)
            reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(y), None)
            std_err_ = xp.exp(time_pd * reg.slope_ + reg.intercept_)

            self.sigma_.values = xp.concatenate(
                (self.sigma_.values[..., :-1], sigma_[..., -1:]), axis=-1
            )
            self.std_err_.values = xp.concatenate(
                (self.std_err_.values[..., :-1], std_err_[..., -1:]), axis=-1
            )

    def _get_tail_weighted_time_period(self, X):
        """ Method to approximate the weighted-average development age of tail
        using log-linear extrapolation

        Returns: float32
        """
        y = X.ldf_.values.copy()
        xp = X.ldf_.get_array_module()
        y[y <= 1] = xp.nan
        reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(y - 1), None)
        tail = xp.prod(
            self.ldf_.values[..., -self._ave_period[0] - 1 :], -1, keepdims=True
        )
        reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(y - 1), None)
        tail = tail if tail.max() > 1 else 1.001
        time_pd = (xp.log(tail - 1) - reg.intercept_) / reg.slope_
        return time_pd

    @staticmethod
    def _tail_(self):
        df = self.cdf_[
            self.cdf_.development
            == self.cdf_.development.iloc[-1 - self._ave_period[0]]
        ]
        if np.all(df.values.min(axis=2) == df.values.max(axis=2)):
            df = df.iloc[..., 0, :].to_frame()
        return df

    @property
    def tail_(self):
        return TailBase._tail_(self)



import chainladder as cl
import pandas as pd
import numpy as np
import copy

tri = cl.load_sample("clrd")
qtr = cl.load_sample("quarterly")
raa = cl.load_sample("raa")
tri_gt = copy.deepcopy(tri)
qtr_gt = copy.deepcopy(qtr)
raa_gt = copy.deepcopy(raa)
# Test Triangle slicing
def test_slice_by_boolean():
    assert tri == tri_gt
    assert (
        tri[tri["LOB"] == "ppauto"].loc["Wolverine Mut Ins Co"]["CumPaidLoss"]
        == tri.loc["Wolverine Mut Ins Co"].loc["ppauto"]["CumPaidLoss"]
    )


def test_slice_by_loc():
    assert tri == tri_gt
    assert tri.loc["Aegis Grp"].loc["comauto"].index.iloc[0, 0] == "comauto"


def test_slice_origin():
    assert raa == raa_gt
    assert raa[raa.origin > "1985"].shape == (1, 1, 5, 10)


def test_slice_development():
    assert raa == raa_gt
    assert raa[raa.development < 72].shape == (1, 1, 10, 5)


def test_slice_by_loc_iloc():
    assert tri == tri_gt
    assert tri.groupby("LOB").sum().loc["comauto"].index.iloc[0, 0] == "comauto"


def test_repr():
    assert raa == raa_gt
    np.testing.assert_array_equal(
        pd.read_html(raa._repr_html_())[0].set_index("Unnamed: 0").values,
        raa.to_frame().values,
    )


def test_arithmetic_union():
    assert raa == raa_gt
    assert raa.shape == (raa - raa[raa.valuation < "1987"]).shape


def test_to_frame_unusual():
    assert tri == tri_gt
    a = tri.groupby(["LOB"]).sum().latest_diagonal["CumPaidLoss"].to_frame().values
    b = tri.latest_diagonal["CumPaidLoss"].groupby(["LOB"]).sum().to_frame().values
    np.testing.assert_array_equal(a, b)


def test_link_ratio():
    assert raa == raa_gt
    xp = raa.get_array_module()
    assert (
        xp.sum(
            xp.nan_to_num(raa.link_ratio.values * raa.values[:, :, :-1, :-1])
            - xp.nan_to_num(raa.values[:, :, :-1, 1:])
        )
        < 1e-5
    )


def test_incr_to_cum():
    assert tri == tri_gt
    xp = tri.get_array_module()
    xp.testing.assert_array_equal(tri.cum_to_incr().incr_to_cum().values, tri.values)


def test_create_new_value():
    assert tri == tri_gt
    tri2 = tri.copy()
    tri2["lr"] = tri2["CumPaidLoss"] / tri2["EarnedPremDIR"]
    assert (tri.shape[0], tri.shape[1] + 1, tri.shape[2], tri.shape[3]) == tri2.shape


def test_multilevel_index_groupby_sum1():
    assert tri == tri_gt
    assert tri.groupby("LOB").sum().sum() == tri.sum()


def test_multilevel_index_groupby_sum2():
    assert tri == tri_gt
    a = tri.groupby("GRNAME").sum().sum()
    b = tri.groupby("LOB").sum().sum()
    assert a == b


def test_boolean_groupby_eq_groupby_loc():
    assert tri == tri_gt
    xp = tri.get_array_module()
    xp.testing.assert_array_equal(
        tri[tri["LOB"] == "ppauto"].sum().values,
        tri.groupby("LOB").sum().loc["ppauto"].values,
    )


def test_latest_diagonal_two_routes():
    assert tri == tri_gt
    assert (
        tri.latest_diagonal.sum()["BulkLoss"] == tri.sum().latest_diagonal["BulkLoss"]
    )


def test_sum_of_diff_eq_diff_of_sum():
    assert tri == tri_gt
    assert (tri["BulkLoss"] - tri["CumPaidLoss"]).latest_diagonal == (
        tri.latest_diagonal["BulkLoss"] - tri.latest_diagonal["CumPaidLoss"]
    )


def test_append():
    assert raa == raa_gt
    raa2 = raa.copy()
    raa2.kdims = np.array([["P2"]])
    raa.append(raa2).sum() == raa * 2
    assert raa.append(raa2).sum() == 2 * raa


def test_assign_existing_col():
    qtr = cl.load_sample("quarterly")
    before = qtr.shape
    qtr["paid"] = 1 / qtr["paid"]
    assert qtr.shape == before


def test_arithmetic_across_keys():
    x = cl.load_sample("auto")
    xp = x.get_array_module()
    xp.testing.assert_array_equal((x.sum() - x.iloc[0]).values, x.iloc[1].values)


def test_grain():
    assert qtr == qtr_gt
    actual = qtr.iloc[0, 0].grain("OYDY")
    xp = actual.get_array_module()
    expected = xp.array(
        [
            [
                44.0,
                621.0,
                950.0,
                1020.0,
                1070.0,
                1069.0,
                1089.0,
                1094.0,
                1097.0,
                1099.0,
                1100.0,
                1100.0,
            ],
            [
                42.0,
                541.0,
                1052.0,
                1169.0,
                1238.0,
                1249.0,
                1266.0,
                1269.0,
                1296.0,
                1300.0,
                1300.0,
                xp.nan,
            ],
            [
                17.0,
                530.0,
                966.0,
                1064.0,
                1100.0,
                1128.0,
                1155.0,
                1196.0,
                1201.0,
                1200.0,
                xp.nan,
                xp.nan,
            ],
            [
                10.0,
                393.0,
                935.0,
                1062.0,
                1126.0,
                1209.0,
                1243.0,
                1286.0,
                1298.0,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                13.0,
                481.0,
                1021.0,
                1267.0,
                1400.0,
                1476.0,
                1550.0,
                1583.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                2.0,
                380.0,
                788.0,
                953.0,
                1001.0,
                1030.0,
                1066.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                4.0,
                777.0,
                1063.0,
                1307.0,
                1362.0,
                1411.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                2.0,
                472.0,
                1617.0,
                1818.0,
                1820.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                3.0,
                597.0,
                1092.0,
                1221.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                4.0,
                583.0,
                1212.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                21.0,
                422.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
            [
                13.0,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
                xp.nan,
            ],
        ]
    )
    xp.testing.assert_array_equal(actual.values[0, 0, :, :], expected)


def test_off_cycle_val_date():
    assert qtr == qtr_gt
    assert qtr.valuation_date.strftime("%Y-%m-%d") == "2006-03-31"


def test_printer():
    print(cl.load_sample("abc"))


def test_value_order():
    assert tri == tri_gt
    a = tri[["CumPaidLoss", "BulkLoss"]]
    b = tri[["BulkLoss", "CumPaidLoss"]]
    xp = a.get_array_module()
    xp.testing.assert_array_equal(a.values[:, -1], b.values[:, 0])


def test_trend():
    assert (
        abs(
            (
                cl.load_sample("abc").trend(0.05).trend((1 / 1.05) - 1)
                - cl.load_sample("abc")
            )
            .sum()
            .sum()
        )
        < 1e-5
    )


def test_arithmetic_1():
    x = cl.load_sample("mortgage")
    np.testing.assert_array_equal(-(((x / x) + 0) * x), -(+x))


def test_arithmetic_2():
    x = cl.load_sample("mortgage")
    np.testing.assert_array_equal(1 - (x / x), 0 * x * 0)


def test_rtruediv():
    assert raa == raa_gt
    xp = raa.get_array_module()
    assert xp.nansum(abs(((1 / raa) * raa).values[0, 0] - raa.nan_triangle)) < 0.00001


def test_shift():
    assert qtr == qtr_gt
    x = qtr.iloc[0, 0]
    xp = x.get_array_module()
    xp.testing.assert_array_equal(x[x.valuation <= x.valuation_date].values, x.values)


def test_quantile_vs_median():
    assert tri == tri_gt
    xp = tri.get_array_module()
    xp.testing.assert_array_equal(
        tri.quantile(q=0.5)["CumPaidLoss"].values, tri.median()["CumPaidLoss"].values
    )


def test_grain_returns_valid_tri():
    assert qtr == qtr_gt
    assert qtr.grain("OYDY").latest_diagonal == qtr.latest_diagonal


def test_base_minimum_exposure_triangle():
    assert raa == raa_gt
    d = (raa.latest_diagonal * 0 + 50000).to_frame().reset_index()
    d["index"] = d["index"].astype(str)
    cl.Triangle(d, origin="index", columns=d.columns[-1])


def test_origin_and_value_setters():
    assert raa == raa_gt
    raa2 = raa.copy()
    raa.columns = list(raa.columns)
    raa.origin = list(raa.origin)
    assert np.all(
        (
            np.all(raa2.origin == raa.origin),
            np.all(raa2.development == raa.development),
            np.all(raa2.odims == raa.odims),
            np.all(raa2.vdims == raa.vdims),
        )
    )


def test_grain_increm_arg():
    assert qtr == qtr_gt
    tri_i = qtr["incurred"].cum_to_incr()
    a = tri_i.grain("OYDY").incr_to_cum()
    assert a == qtr["incurred"].grain("OYDY").set_backend(a.array_backend)


def test_valdev1():
    assert qtr == qtr_gt
    a = qtr.dev_to_val().val_to_dev()
    b = qtr
    assert a == b


def test_valdev2():
    assert qtr == qtr_gt
    a = qtr.dev_to_val().grain("OYDY").val_to_dev()
    b = qtr.grain("OYDY")
    assert a == b


def test_valdev3():
    assert qtr == qtr_gt
    a = qtr.grain("OYDY").dev_to_val().val_to_dev()
    b = qtr.grain("OYDY")
    assert a == b


# def test_valdev4():
#    # Does not work with pandas 0.23, consider requiring only pandas>=0.24
#    raa = raa
#    np.testing.assert_array_equal(raa.dev_to_val()[raa.dev_to_val().development>='1989'].values,
#        raa[raa.valuation>='1989'].dev_to_val().values)


def test_valdev5():
    assert raa == raa_gt
    xp = raa.get_array_module()
    xp.testing.assert_array_equal(
        raa[raa.valuation >= "1989"].latest_diagonal.values, raa.latest_diagonal.values
    )


def test_valdev6():
    assert raa == raa_gt
    xp = raa.get_array_module()
    xp.testing.assert_array_equal(
        raa.grain("OYDY").latest_diagonal.values,
        raa.latest_diagonal.grain("OYDY").values,
    )


def test_valdev7():
    assert qtr == qtr_gt
    xp = qtr.get_array_module()
    x = cl.Chainladder().fit(qtr).full_expectation_
    assert xp.sum(x.dev_to_val().val_to_dev().values - x.values) < 1e-5


def test_reassignment():
    assert tri == tri_gt
    clrd = tri.copy()
    clrd["values"] = clrd["CumPaidLoss"]
    clrd["values"] = clrd["values"] + clrd["CumPaidLoss"]


def test_dropna():
    assert tri == tri_gt
    assert tri.shape == tri.dropna().shape
    a = tri[tri["LOB"] == "wkcomp"].iloc[-5]["CumPaidLoss"].dropna().shape
    assert a == (1, 1, 2, 2)


def test_commutative():
    assert qtr == qtr_gt
    xp = qtr.get_array_module()
    full = cl.Chainladder().fit(qtr).full_expectation_
    assert qtr.grain("OYDY").val_to_dev() == qtr.val_to_dev().grain("OYDY")
    assert qtr.cum_to_incr().grain(
        "OYDY"
    ).val_to_dev() == qtr.val_to_dev().cum_to_incr().grain("OYDY")
    assert qtr.grain(
        "OYDY"
    ).cum_to_incr().val_to_dev().incr_to_cum() == qtr.val_to_dev().grain("OYDY")
    assert full.grain("OYDY").val_to_dev() == full.val_to_dev().grain("OYDY")
    assert full.cum_to_incr().grain(
        "OYDY"
    ).val_to_dev() == full.val_to_dev().cum_to_incr().grain("OYDY")
    a = full.grain("OYDY").cum_to_incr().val_to_dev().incr_to_cum()
    b = full.val_to_dev().grain("OYDY")
    assert abs(a - b).max().max().max() < 1e-5


def test_broadcasting():
    assert tri == tri_gt
    assert raa == raa_gt
    t1 = raa
    t2 = tri
    assert t1.broadcast_axis("columns", t2.columns).shape[1] == t2.shape[1]
    assert t1.broadcast_axis("index", t2.index).shape[0] == t2.shape[0]
    raa.latest_diagonal.to_frame()


def test_slicers_honor_order():
    assert tri == tri_gt
    clrd = tri.groupby("LOB").sum()
    assert clrd.iloc[[1, 0], :].iloc[0, 1] == clrd.iloc[1, 1]  # row
    assert clrd.iloc[[1, 0], [1, 0]].iloc[0, 0] == clrd.iloc[1, 1]  # col
    assert clrd.loc[:, ["CumPaidLoss", "IncurLoss"]].iloc[0, 0] == clrd.iloc[0, 1]
    assert (
        clrd.loc[["ppauto", "medmal"], ["CumPaidLoss", "IncurLoss"]].iloc[0, 0]
        == clrd.iloc[3]["CumPaidLoss"]
    )
    assert (
        clrd.loc[clrd["LOB"] == "comauto", ["CumPaidLoss", "IncurLoss"]]
        == clrd[clrd["LOB"] == "comauto"].iloc[:, [1, 0]]
    )
    assert clrd.groupby("LOB").sum() == clrd


def test_exposure_tri():
    x = cl.load_sample("auto")
    x = x[x.development == 12]
    x = x["paid"].to_frame().T.unstack().reset_index()
    x.columns = ["LOB", "origin", "paid"]
    x.origin = x.origin.astype(str)
    y = cl.Triangle(x, origin="origin", index="LOB", columns="paid")
    x = cl.load_sample("auto")["paid"]
    x = x[x.development == 12]
    assert x == y


def test_jagged_1_add():
    assert raa == raa_gt
    raa1 = raa[raa.origin <= "1984"]
    raa2 = raa[raa.origin > "1984"]
    assert raa2 + raa1 == raa
    assert raa2.dropna() + raa1.dropna() == raa


def test_jagged_2_add():
    assert raa == raa_gt
    raa1 = raa[raa.development <= 48]
    raa2 = raa[raa.development > 48]
    assert raa2 + raa1 == raa
    assert raa2.dropna() + raa1.dropna() == raa


def test_df_period_input():
    assert raa == raa_gt
    d = raa.latest_diagonal
    df = d.to_frame().reset_index()
    assert cl.Triangle(df, origin="index", columns=df.columns[-1]) == d


def test_trend_on_vector():
    assert raa == raa_gt
    d = raa.latest_diagonal
    assert d.trend(0.05, axis=2).to_frame().astype(int).iloc[0, 0] == 29217


def test_latest_diagonal_val_to_dev():
    assert raa == raa_gt
    assert raa.latest_diagonal.val_to_dev() == raa[raa.valuation == raa.valuation_date]


def test_vector_division():
    assert raa == raa_gt
    raa.latest_diagonal / raa


def test_sumdiff_to_diffsum():
    assert tri == tri_gt
    out = tri["CumPaidLoss"]
    assert out.cum_to_incr().incr_to_cum().sum() == out.sum()


def test_multiindex_broadcast():
    assert tri == tri_gt
    clrd = tri["CumPaidLoss"]
    clrd / clrd.groupby("LOB").sum()


def test_backends():
    assert tri == tri_gt
    clrd = tri[["CumPaidLoss", "EarnedPremDIR"]]
    a = clrd.iloc[1, 0].set_backend("sparse").dropna()
    b = clrd.iloc[1, 0].dropna()
    assert a == b


def test_union_columns():
    assert tri == tri_gt
    assert tri.iloc[:, :3] + tri.iloc[:, 3:] == tri


def test_4loc():
    assert tri == tri_gt
    clrd = tri.groupby("LOB").sum()
    assert (
        clrd.iloc[:3, :2, 0, 0]
        == clrd[clrd.origin == tri.origin.min()][
            clrd.development == clrd.development.min()
        ].loc["comauto":"othliab", :"CumPaidLoss", :, :]
    )
    assert (
        clrd.iloc[:3, :2, 0:1, -1]
        == clrd[clrd.development == tri.development.max()].loc[
            "comauto":"othliab", :"CumPaidLoss", "1988", :
        ]
    )


def test_init_vector():
    assert raa == raa_gt
    a = raa[raa.development == 12]
    b = pd.DataFrame(
        {"AccYear": [item for item in range(1981, 1991)], "premium": [3000000] * 10}
    )
    b = cl.Triangle(b, origin="AccYear", columns="premium")
    assert np.all(a.valuation == b.valuation)
    assert a.valuation_date == b.valuation_date


def test_loc_ellipsis():
    assert tri == tri_gt
    assert (
        tri.loc["Aegis Grp"] == tri.loc["Adriatic Ins Co":"Aegis Grp"].loc["Aegis Grp"]
    )
    assert tri.loc["Aegis Grp", ..., :] == tri.loc["Aegis Grp"]
    assert tri.loc[..., 24:] == tri.loc[..., :, 24:]
    assert tri.loc[:, ..., 24:] == tri.loc[..., :, 24:]
    assert tri.loc[:, "CumPaidLoss"] == tri.loc[:, "CumPaidLoss", ...]
    assert tri.loc[..., "CumPaidLoss", :, :] == tri.loc[:, "CumPaidLoss", :, :]


def missing_first_lag():
    x = raa.copy()
    x.values[:, :, :, 0] = 0
    x = x.sum(0)
    x.link_ratio.shape == (1, 1, 9, 9)


def test_reverse_slice_integrity():
    assert tri == tri_gt
    assert tri.iloc[::-1, ::-1].shape == tri.shape
    assert np.all(tri.iloc[:, ::-1].columns.values == tri.columns[::-1])
    assert tri.iloc[tri.index.index[::-1]] + tri == 2 * tri.iloc[::-1]


def test_loc_tuple():
    assert tri == tri_gt
    assert len(tri.loc[("Adriatic Ins Co", "othliab")]) == 1
    assert tri.loc[tri.index] == tri


def test_index_broadcasting():
    assert tri == tri_gt
    assert ((tri / tri.sum()) - ((1 / tri.sum()) * tri)).sum().sum().sum() < 1e-4


def test_groupby_axis1():
    assert tri == tri_gt
    clrd = tri.sum("origin").sum("development")
    groups = [i.find("Loss") >= 0 for i in clrd.columns]
    assert np.all(
        clrd.to_frame().groupby(groups, axis=1).sum()
        == clrd.groupby(groups, axis=1).sum().to_frame()
    )
    assert np.all(
        clrd.to_frame().groupby("LOB").sum() == clrd.groupby("LOB").sum().to_frame()
    )
    cl.load_sample("prism").sum().grain("OYDY")


def test_different_forms_of_grain():
    t = cl.load_sample("prism").sum()["Paid"]
    assert (
        abs(t.grain("OYDY") - t.incr_to_cum().grain("OYDY").cum_to_incr()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.grain("OYDY", trailing=True)
            - t.incr_to_cum().grain("OYDY", trailing=True).cum_to_incr()
        )
        .sum()
        .sum()
        < 1e-4
    )
    assert (
        abs(t.incr_to_cum().grain("OYDY") - t.grain("OYDY").incr_to_cum()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.incr_to_cum().grain("OYDY", trailing=True)
            - t.grain("OYDY", trailing=True).incr_to_cum()
        )
        .sum()
        .sum()
        < 1e-4
    )
    assert (
        abs(t.grain("OYDQ") - t.incr_to_cum().grain("OYDQ").cum_to_incr()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.grain("OYDQ", trailing=True)
            - t.incr_to_cum().grain("OYDQ", trailing=True).cum_to_incr()
        )
        .sum()
        .sum()
        < 1e-4
    )
    assert (
        abs(t.incr_to_cum().grain("OQDM") - t.grain("OQDM").incr_to_cum()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.incr_to_cum().grain("OQDM", trailing=True)
            - t.grain("OQDM", trailing=True).incr_to_cum()
        )
        .sum()
        .sum()
        < 1e-4
    )
    t = t.dev_to_val()
    assert (
        abs(t.grain("OYDY") - t.incr_to_cum().grain("OYDY").cum_to_incr()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.grain("OYDY", trailing=True)
            - t.incr_to_cum().grain("OYDY", trailing=True).cum_to_incr()
        )
        .sum()
        .sum()
        < 1e-4
    )
    assert (
        abs(t.incr_to_cum().grain("OYDY") - t.grain("OYDY").incr_to_cum()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.incr_to_cum().grain("OYDY", trailing=True)
            - t.grain("OYDY", trailing=True).incr_to_cum()
        )
        .sum()
        .sum()
        < 1e-4
    )
    t = t.val_to_dev()
    t = t[t.valuation < "2017-09"]
    assert (
        abs(t.grain("OYDY") - t.incr_to_cum().grain("OYDY").cum_to_incr()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.grain("OYDY", trailing=True)
            - t.incr_to_cum().grain("OYDY", trailing=True).cum_to_incr()
        )
        .sum()
        .sum()
        < 1e-4
    )
    assert (
        abs(t.incr_to_cum().grain("OYDY") - t.grain("OYDY").incr_to_cum()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.incr_to_cum().grain("OYDY", trailing=True)
            - t.grain("OYDY", trailing=True).incr_to_cum()
        )
        .sum()
        .sum()
        < 1e-4
    )
    t = t.dev_to_val()
    assert (
        abs(t.grain("OYDY") - t.incr_to_cum().grain("OYDY").cum_to_incr()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.grain("OYDY", trailing=True)
            - t.incr_to_cum().grain("OYDY", trailing=True).cum_to_incr()
        )
        .sum()
        .sum()
        < 1e-4
    )
    assert (
        abs(t.incr_to_cum().grain("OYDY") - t.grain("OYDY").incr_to_cum()).sum().sum()
        < 1e-4
    )
    assert (
        abs(
            t.incr_to_cum().grain("OYDY", trailing=True)
            - t.grain("OYDY", trailing=True).incr_to_cum()
        )
        .sum()
        .sum()
        < 1e-4
    )


def test_partial_year():
    before = cl.load_sample('prism')['Paid'].sum().incr_to_cum()
    before=before[before.valuation<='2017-08'].latest_diagonal

    after = cl.Triangle(
        before.to_frame(keepdims=True).reset_index(),
        origin='origin', development='valuation', columns='Paid', index=before.key_labels)

    assert after.valuation_date == before.valuation_date


def test_at_iat():
    raa1 = cl.load_sample('raa')
    raa2 = cl.load_sample('raa')
    raa1.at['Total','values', '1985', 120] = 5
    raa1.at['Total','values', '1985', 12] = 5
    raa2.iat[0, 0, 4, -1] = 5
    raa2.iat[-1, -1, 4, 0] = 5
    assert raa1 == raa2


def test_at_iat_sparse():
    raa1 = cl.load_sample('raa').set_backend('sparse')
    raa2 = cl.load_sample('raa').set_backend('sparse')
    raa1.at['Total','values', '1985', 120] = 5
    raa1.at['Total','values', '1985', 12] = 5
    raa2.iat[0, 0, 4, -1] = 5
    raa2.iat[-1, -1, 4, 0] = 5
    assert raa1 == raa2



""" utils should store all utility functions and classes, i.e. things that
    are used by various modules in the package.
"""
from sklearn.utils import deprecated
from chainladder.utils.weighted_regression import (
    WeightedRegression,
)  # noqa (API import)
from xlcompose import (
    DataFrame,
    Series,
    Row,
    Column,
    Tabs,
    CSpacer,
    RSpacer,
    Title,
    Image,
    VSpacer,
    HSpacer,
    Sheet,
    load_yaml,
)
import os

@deprecated('The load_template class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')
def load_template(template, env=None, **kwargs):
    path = os.path.dirname(os.path.abspath(__file__))
    try:
        return load_yaml(template, env, **kwargs)
    except:
        template = os.path.join(path, "templates", template.lower() + ".yaml")
        return load_yaml(template, env, **kwargs)

DatFrame = deprecated('The DataFrame class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(DataFrame)
Series = deprecated('The Series class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(Series)
Row = deprecated('The Row class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(Row)
Column = deprecated('The Column class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(Column)
Tabs = deprecated('The Tabs class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(Tabs)
CSpacer = deprecated('The CSpacer class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(CSpacer)
RSpacer = deprecated('The RSpacer class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(RSpacer)
Title = deprecated('The Title class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(Title)
Image = deprecated('The Image class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(Image)
VSpacer = deprecated('The VSpacer class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(VSpacer)
HSpacer = deprecated('The HSpacer class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(HSpacer)
Sheet = deprecated('The Sheet class is deprecated and will be removed in a future version of chainladder. Use xlcompose library directly for the same functionality')(Sheet)
load_template = (load_template)

from chainladder.utils.utility_functions import (  # noqa (API import)
    load_dataset,
    parallelogram_olf,
    read_pickle,
    read_json,
    concat,
    load_sample,
    minimum,
    maximum,
    PatsyFormula,
)
from chainladder.utils.cupy import cp
from chainladder.utils.sparse import sp



ARRAY_BACKEND = "numpy"
AUTO_SPARSE = True
ARRAY_PRIORITY = ["sparse", "cupy", "numpy"]
ULT_VAL = "2261-12-31 23:59:59.999999999"


def array_backend(array_backend="numpy"):
    global ARRAY_BACKEND
    ARRAY_BACKEND = array_backend


def auto_sparse(auto_sparse=True):
    global AUTO_SPARSE
    AUTO_SPARSE = auto_sparse


from chainladder.utils import *  # noqa (API Import)
from chainladder.core import *  # noqa (API Import)
from chainladder.development import *  # noqa (API Import)
from chainladder.adjustments import *  # noqa (API Import)
from chainladder.tails import *  # noqa (API Import)
from chainladder.methods import *  # noqa (API Import)
from chainladder.workflow import *  # noqa (API Import)

__version__ = "0.8.3"



from chainladder.development.base import Development, DevelopmentBase
from chainladder import ULT_VAL
from chainladder.utils.utility_functions import concat

import numpy as np
import pandas as pd


class CaseOutstanding(DevelopmentBase):
    """ A determinisic method based on outstanding case reserves.

    The CaseOutstanding method is a deterministic approach that develops
    patterns of incremental payments as a percent of previous period case
    reserves as well as patterns for case reserves as a percent of previous
    period case reserves.  Although the patterns produces by the approach
    approximate incremental payments and case outstanding, they are converted
    into comparable multiplicative patterns for usage with the various IBNR
    methods.

    .. versionadded:: 0.8.0

    Parameters
    ----------
    paid_to_incurred : tuple or list of tuples
        A tuple representing the paid and incurred ``columns`` of the triangles
        such as ``('paid', 'incurred')``
    paid_n_periods : integer, optional (default=-1)
        number of origin periods to be used in the paid pattern averages. For
        all origin periods, set paid_n_periods=-1
    case_n_periods : integer, optional (default=-1)
        number of origin periods to be used in the case pattern averages. For
        all origin periods, set paid_n_periods=-1

    Attributes
    ----------
    ldf_ : Triangle
        The estimated (multiplicative) loss development patterns.
    cdf_ : Triangle
        The estimated (multiplicative) cumulative development patterns.
    case_to_prior_case_ : Triangle
        The case to prior case ratios used for fitting the estimator
    case_ldf_ :
        The selected case to prior case ratios of the fitted estimator
    paid_to_prior_case_ : Triangle
        The paid to prior case ratios used for fitting the estimator
    paid_ldf_ :
        The selected paid to prior case ratios of the fitted estimator
    """
    def __init__(self, paid_to_incurred=None, paid_n_periods=-1, case_n_periods=-1):
        self.paid_to_incurred = paid_to_incurred
        self.paid_n_periods = paid_n_periods
        self.case_n_periods = case_n_periods

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        backend = "cupy" if X.array_backend == "cupy" else "numpy"
        self.X_ = X.copy()
        paid_tri = self.X_[self.paid_to_incurred[0]]
        incurred_tri = self.X_[self.paid_to_incurred[1]]
        self.paid_w_ = Development(n_periods=self.paid_n_periods).fit(self.X_.iloc[0,0]).w_
        self.case_w_ = Development(n_periods=self.case_n_periods).fit(self.X_.iloc[0,0]).w_
        self.case_ldf_ = self.case_to_prior_case_.mean(2)
        self.paid_ldf_ = self.paid_to_prior_case_.mean(2)

        case = incurred_tri-paid_tri
        patterns = ((1 - np.nan_to_num(case.nan_triangle[..., 1:])) *
                    self.case_ldf_.values)
        for i in range(np.isnan(case.nan_triangle[-1]).sum()):
            increment = (
                (case - case[case.valuation < case.valuation_date]).iloc[..., :-1] *
                patterns)
            increment.ddims = case.ddims[1:]
            increment.valuation_date = case.valuation[case.valuation>=case.valuation_date].drop_duplicates()[1]
            case = case + increment

        patterns = ((1-np.nan_to_num(self.X_.nan_triangle[..., 1:]))*self.paid_ldf_.values)

        paid = (case.iloc[..., :-1]*patterns)
        paid.ddims = case.ddims[1:]
        paid.valuation_date = pd.Timestamp(ULT_VAL)
        paid = (paid_tri.cum_to_incr() + paid).incr_to_cum()
        inc = (case[case.valuation>self.X_.valuation_date] +
               paid[paid.valuation>self.X_.valuation_date] +
               incurred_tri)
        paid.columns = [self.paid_to_incurred[0]]
        inc.columns = [self.paid_to_incurred[1]]
        cols = self.X_.columns[self.X_.columns.isin([self.paid_to_incurred[0], self.paid_to_incurred[1]])]
        dev = concat((paid, inc), 1)[list(cols)]
        self.dev_ = dev
        dev = (dev.iloc[..., -1]/dev).iloc[..., :-1]
        dev.valuation_date = pd.Timestamp(ULT_VAL)
        dev.ddims = self.X_.link_ratio.ddims
        dev.is_pattern=True
        dev.is_cumulative=True

        self.ldf_ = dev.cum_to_incr().set_backend(backend)
        return self

    @property
    def case_to_prior_case_(self):
        paid_tri = self.X_[self.paid_to_incurred[0]]
        incurred_tri = self.X_[self.paid_to_incurred[1]]
        out = ((incurred_tri - paid_tri).iloc[..., 1:] * self.case_w_ /
               (incurred_tri - paid_tri).iloc[..., :-1].values)
        out.is_pattern = True
        out.is_cumulative=False
        return out

    @property
    def paid_to_prior_case_(self):
        paid_tri = self.X_[self.paid_to_incurred[0]]
        incurred_tri = self.X_[self.paid_to_incurred[1]]
        out = (
            paid_tri.cum_to_incr().iloc[..., 1:] * self.paid_w_ /
            (incurred_tri - paid_tri).iloc[..., :-1].values)
        out.is_pattern=True
        return out

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        triangles = ["ldf_"]
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.methods import Benktander
import warnings


class CapeCod(Benktander):
    """Applies the CapeCod technique to triangle **X**

    Parameters
    ----------
    trend : float (default=0.0)
        The cape cod trend assumption.  Any Trend transformer on X will override
        this argument.
    decay : float (defaut=1.0)
        The cape cod decay assumption
    n_iters : int, optional (default=1)
        Number of iterations to use in the Benktander model.
    apriori_sigma : float, optional (default=0.0)
        Standard deviation of the apriori.  When used in conjunction with the
        bootstrap model, the model samples aprioris from a lognormal distribution
        using this argument as a standard deviation.
    random_state : int, RandomState instance or None, optional (default=None)
        Seed for sampling from the apriori distribution.  This is ignored when
        using as a deterministic method.
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random.


    Attributes
    ----------
    triangle :
        returns **X**
    ultimate_ :
        The ultimate losses per the method
    ibnr_ :
        The IBNR per the method
    apriori_ :
        The trended apriori vector developed by the Cape Cod Method
    detrended_apriori_ :
        The detrended apriori vector developed by the Cape Cod Method
    """

    def __init__(self, trend=0, decay=1, n_iters=1, apriori_sigma=0.0, random_state=None):
        self.trend = trend
        self.decay = decay
        self.n_iters = n_iters
        self.apriori_sigma = apriori_sigma
        self.random_state = random_state

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Loss data to which the model will be applied.
        y : None
            Ignored
        sample_weight : Triangle-like
            The exposure to be used in the method.
        Returns
        -------
        self : object
            Returns the instance itself.
        """

        if sample_weight is None:
            raise ValueError("sample_weight is required.")
        self.apriori = 1.0
        self.X_ = self.validate_X(X)
        self.validate_weight(X, sample_weight)
        sample_weight = sample_weight.set_backend(self.X_.array_backend)
        self.apriori_, self.detrended_apriori_ = self._get_capecod_aprioris(X, sample_weight)
        self.expectation_ = self.detrended_apriori_ * sample_weight.values
        super().fit(X, y, self.expectation_)
        return self

    def _get_capecod_aprioris(self, X, sample_weight):
        """ Private method to establish CapeCod Apriori """
        xp = X.get_array_module()
        ult = X.copy()
        latest = X.latest_diagonal.values
        len_orig = sample_weight.shape[-2]
        reported_exposure = sample_weight.values / self._align_cdf(ult, sample_weight)
        if hasattr(X, "trend_"):
            if self.trend != 0:
                warnings.warn(
                    "CapeCod Trend assumption is ignored when X has a trend_ property."
                )
            trend_array = X.trend_.latest_diagonal.values
        else:
            trend_array = (X.trend(self.trend) / X).latest_diagonal.values
        if hasattr(sample_weight, "olf_"):
            sw_olf_array = sample_weight.olf_.values
        else:
            sw_olf_array = 1
        if hasattr(X, "olf_"):
            X_olf_array = X.olf_.values
        else:
            X_olf_array = 1
        decay_matrix = self.decay ** xp.abs(
            xp.arange(len_orig)[None].T - xp.arange(len_orig)[None]
        )
        weighted_exposure = xp.swapaxes(reported_exposure, -1, -2) * decay_matrix
        trended_ultimate = (latest * trend_array * X_olf_array) / (
            reported_exposure * sw_olf_array
        )
        trended_ultimate = xp.swapaxes(trended_ultimate, -1, -2)
        apriori = xp.sum(weighted_exposure * trended_ultimate, -1) / xp.sum(
            weighted_exposure, -1
        )
        ult.values = apriori[..., None]

        apriori_ = ult.copy()
        detrended_ultimate = apriori_.values / trend_array / X_olf_array * sw_olf_array
        detrended_apriori_ = ult.copy()
        detrended_apriori_.values = detrended_ultimate
        return self._set_ult_attr(apriori_), self._set_ult_attr(detrended_apriori_)


    def predict(self, X, sample_weight=None):
        """Predicts the CapeCod ultimate on a new triangle **X**

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.
        sample_weight : Triangle
            For exposure-based methods, the exposure to be used for predictions

        Returns
        -------
        X_new: Triangle
            Loss data with CapeCod ultimate applied
        """
        if sample_weight is None:
            raise ValueError("sample_weight is required.")
        apriori_, detrended_apriori_ = self._get_capecod_aprioris(
            X, sample_weight
        )
        expectation_ = detrended_apriori_ * sample_weight.values
        X_new = super().predict(X,  expectation_)
        X_new.apriori_ = apriori_
        X_new.detrended_apriori_ = detrended_apriori_
        X_new.expectation_ = expectation_
        return X_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.development.base import Development, DevelopmentBase
import numpy as np
import pandas as pd
import warnings


class IncrementalAdditive(DevelopmentBase):
    """ The Incremental Additive Method.

    Parameters
    ----------
    trend : float (default=0.0)
        A multiplicative trend amount used to trend each incremental development
        period the valuation_date of the Triangle.
    future_trend : float (default=None)
        The trend to apply to the incremental development periods in the lower
        half of the completed Triangle. If None, then will be set to the value of
        the trend parameter.
    n_periods : integer, optional (default=-1)
        number of origin periods to be used in the ldf average calculation. For
        all origin periods, set n_periods=-1
    average: str optional (default='volume')
        type of averaging to use for ldf average calculation.  Options include
        'volume' and 'simple'.
    drop : tuple or list of tuples
        Drops specific origin/development combination(s)
    drop_high : bool or list of bool (default=None)
        Drops highest link ratio(s) from LDF calculation
    drop_low : bool or list of bool (default=None)
        Drops lowest link ratio(s) from LDF calculation
    drop_valuation : str or list of str (default=None)
        Drops specific valuation periods. str must be date convertible.

    Attributes
    ----------
    ldf_ : Triangle
        The estimated loss development patterns
    cdf_ : Triangle
        The estimated cumulative development patterns
    zeta : Triangle
        The fitted incrementals as a percent of exposure trended to the valuation
        date of the Triangle.
    incremental_ : Triangle
        A triangle of full incremental values.


    """

    def __init__(
        self, trend=0.0, n_periods=-1, average="volume", future_trend=0,
        drop=None, drop_high=None, drop_low=None, drop_valuation=None):
        self.trend = trend
        self.n_periods = n_periods
        self.average = average
        self.future_trend = future_trend
        self.drop_high = drop_high
        self.drop_low = drop_low
        self.drop_valuation = drop_valuation
        self.drop = drop


    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Triangle to which the incremental method is applied.  Triangle must
            be cumulative.
        y : None
            Ignored
        sample_weight :
            Exposure used in the method.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        from chainladder import ULT_VAL
        from chainladder.utils.utility_functions import num_to_nan

        if type(X.ddims) != np.ndarray:
            raise ValueError("Triangle must be expressed with development lags")
        if X.array_backend == "sparse":
            X = X.set_backend("numpy")
        else:
            X = X.copy()
        if sample_weight.array_backend == "sparse":
            sample_weight = sample_weight.set_backend("numpy")
        else:
            sample_weight = sample_weight.copy()
        xp = X.get_array_module()
        sample_weight.is_cumulative = False
        obj = X.cum_to_incr() / sample_weight.values
        if hasattr(X, "trend_"):
            if self.trend != 0:
                warnings.warn(
                    "IncrementalAdditive Trend assumption is ignored when X has a trend_ property."
                )
            x = obj * obj.trend_.values
        else:
            x = obj.trend(self.trend, axis='valuation')

        w_ = Development(
            n_periods=self.n_periods - 1, drop=self.drop,
            drop_high=self.drop_high, drop_low=self.drop_low,
            drop_valuation=self.drop_valuation).fit(x).w_
        # This will miss drops on the latest diagonal
        w_ = num_to_nan(w_)
        w_ = xp.concatenate((w_, (w_[..., -1:] * x.nan_triangle)[..., -1:]), axis=-1)
        if self.average == "simple":
            y_ = xp.nanmean(w_ * x.values, axis=-2)
        if self.average == "volume":
            y_ = xp.nansum(w_ * x.values * sample_weight.values, axis=-2)
            y_ = y_ / xp.nansum(w_ * sample_weight.values, axis=-2)
        self.zeta_ = X.iloc[..., -1:, :]
        self.zeta_.values = y_[:, :, None, :]
        y_ = xp.repeat(y_[..., None, :], len(x.odims), -2)
        obj = x.copy()
        keeps = (
            1
            - xp.nan_to_num(x.nan_triangle)
            + xp.nan_to_num(
                x[x.valuation == x.valuation_date].values[0, 0, ...] * 0 + 1
            )
        )
        obj.values = y_ * keeps
        obj.valuation_date = obj.valuation.max()
        obj.values = obj.values * (1 - xp.nan_to_num(x.nan_triangle)) + xp.nan_to_num(
            (X.cum_to_incr().values / sample_weight.values)
        )

        obj.values[obj.values == 0] = xp.nan
        obj._set_slicers()
        obj.valuation_date = obj.valuation.max()
        future_trend = self.trend if not self.future_trend else self.future_trend
        self.incremental_ = obj * sample_weight.values
        self.incremental_ = self.incremental_.trend(
            1/(1+future_trend)-1, axis='valuation', start=X.valuation_date,
            end=self.incremental_.valuation_date)
        self.ldf_ = obj.incr_to_cum().link_ratio
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        for item in ["ldf_"]:
            X_new.__dict__[item] = self.__dict__[item]
        return X_new
